#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <limits.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/utsname.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <ctype.h>

#include "lib.h"
#include "settings.h"

extern const char *__progname;
extern struct envglobal g;
int detected_libs_32_ = -1;
int __new_glvnd_scheme_detect_result = -1;

void init_global()
{
    g.debug_level = 0;
    g.o_verbose = 0;
    g.other_videocards = 0;
    g.intel_videocards = 0;
    g.nvidia_videocards = 0;
    g.nvidia_3dcontrollers = 0;
    strcpy(g.nv_drv_dir, "");
    strcpy(g.nv_drv32_dir, "");
    strcpy(g.last_kmod_ver_str, "");
    strcpy(g.new_kmod_ver_str, "");
    strcpy(g.proc_kmod_ver_str, "");
    strcpy(g.kernel_ver_str, "");
}

int mk_symlink_from_to(int verbosity, const char *from_file, const char *from_dir, const char *to_file, const char *to_dir)
{
    char src[PATH_MAX];
    char dst[PATH_MAX];
    int need_update = 0;
    int rl_num = 0;
    char link_contents[PATH_MAX];

    strcpy(src, to_dir);
    strcat(src, to_file);
    strcpy(dst, from_dir);
    strcat(dst, from_file);

    /* check symlink before unlink */
    if( (rl_num = readlink(src, link_contents, PATH_MAX-1)) > 0 )
    {
	link_contents[rl_num] = '\0';
	if( strcmp(dst, link_contents) != 0 )
	{
	    if( g.o_verbose )
		fprintf(stderr, "%s: symlink<%s>  need<%s> but link contents<%s>\n", __progname, src, dst, link_contents);
	    need_update = 1;
	}
	else
	{
	    if( (g.o_verbose || verbosity > 0) && verbosity >= 0 )
		printf("%s: don't touch symlink %s (to %s)\n", __progname, src, dst);
	    return 1;
	}
    }
    else
	need_update = 1;

    if( need_update && unlink(src) != 0 )
    {
        if( (g.o_verbose || verbosity > 0) && verbosity >= 0 )
	    fprintf(stderr, "%s: failed to unlink %s: %s\n", __progname, src, strerror(errno));
    }

    if( access(dst, F_OK) != 0 )
    {
	if( (g.o_verbose || verbosity > 0) && verbosity >= 0 )
	    fprintf(stderr, "%s: file %s failed: %s\n", __progname, dst, strerror(errno));
    }

    if( symlink(dst, src) == 0 )
    {
        if( (g.o_verbose || verbosity > 0) && verbosity >= 0 )
	    printf("%s: make symlink %s (to %s)\n", __progname, src, dst);
        return 1;
    }
    else
    {
        if( (g.o_verbose || verbosity > 0) && verbosity >= 0 )
    	    fprintf(stderr, "%s: failed make symlink %s (to %s): %s\n", __progname, src, dst, strerror(errno));
    }
    return 0;
}

int mk_symlink_to(int verbosity, const char *from_file, const char *to_dir, const char *to_file)
{
    return mk_symlink_from_to(verbosity, from_file, g.nv_drv_dir, to_file, to_dir);
}

int mk_symlink32_to(int verbosity, const char *from_file, const char *to_dir, const char *to_file)
{
    return mk_symlink_from_to(verbosity, from_file, g.nv_drv32_dir, to_file, to_dir);
}

int mk_symlink(int verbosity, const char *to_file, const char *to_dir)
{
    return mk_symlink_to(verbosity, to_file, to_dir, to_file);
}

int mk_symlink32(int verbosity, const char *to_file, const char *to_dir)
{
    return mk_symlink32_to(verbosity, to_file, to_dir, to_file);
}

void unlink_file(const char *dir, const char *file)
{
    char dst[PATH_MAX];
    strcpy(dst, dir);
    strcat(dst, file);
    if( unlink(dst) != 0)
    {
	if( g.o_verbose )
	    fprintf(stderr, "%s: failed to unlink %s: %s\n", __progname, dst, strerror(errno));
    }
}


void restore_symlink(int verbosity, const char *from_file, const char *from_dir, const char *to_dir)
{
    char src[PATH_MAX];
    char dst[PATH_MAX];

    strcpy(src, from_dir);
    strcat(src, from_file);

    strcpy(dst, to_dir);
    strcat(dst, from_file);

    if( access(src, F_OK) != 0 )
    {
	if( symlink(dst, src) == 0 )
	{
	    if( (g.o_verbose || verbosity > 0) && verbosity >= 0 )
		printf( "%s: restoring symlink %s (to %s)\n", __progname, src, dst);
	}
	else
	{
	    if( g.o_verbose && verbosity >= 0 )
		fprintf( stderr, "%s: don't restore symlink %s (to %s): %s\n", __progname, src, dst, strerror(errno));
	}
    }
}

void get_kernel_ver()
{
    struct utsname kuname;
    if( uname(&kuname) != 0 )
    {
	fprintf(stderr, "%s: uname(): %s\n", __progname, strerror(errno));
	exit(1);
    }
    if( g.o_verbose )
	printf( "%s: kernel release - %s\n", __progname, kuname.release);
    strcpy(g.kernel_ver_str, kuname.release);
}

void get_last_kmod_ver()
{
    char version_file_path[PATH_MAX];

    strcpy(version_file_path, K_MOD_PREFIX);
    strcat(version_file_path, g.kernel_ver_str);
    strcat(version_file_path, K_MOD_VER_SUFFIX);
    if( g.o_verbose )
	printf( "%s: module version file - %s\n", __progname, version_file_path);

    /* get kernel module version from file */
    if( read_version_from_file(g.last_kmod_ver_str, version_file_path) == 1 )
    {
	if( g.o_verbose )
	    printf( "%s: last kernel module version - %s\n", __progname, g.last_kmod_ver_str );
    }
    else
    {
	strcpy(g.last_kmod_ver_str, "0");
    }
}

void get_nv_drv_dir()
{
    struct stat st;
    strcpy(g.nv_drv_dir, NV_DRV_DIR_PREFIX);
    if( strlen(g.new_kmod_ver_str) > 0 ) {
	strcat(g.nv_drv_dir, g.new_kmod_ver_str);
    } else {
	strcat(g.nv_drv_dir, g.last_kmod_ver_str);
    }
    strcat(g.nv_drv_dir, "/");
    stat(g.nv_drv_dir, &st);
    if( !S_ISDIR(st.st_mode) )
    {
	if( g.o_verbose )
	    fprintf(stderr, "%s: directory %s not found, tring old placement\n", __progname, g.nv_drv_dir);
	strcpy(g.nv_drv_dir, NV_DRV_DIR_PREFIX_OLD);
	if( strlen(g.new_kmod_ver_str) > 0 ) {
	    strcat(g.nv_drv_dir, g.new_kmod_ver_str);
	} else {
	    strcat(g.nv_drv_dir, g.last_kmod_ver_str);
	}
        strcat(g.nv_drv_dir, "/");
    }
    if( g.o_verbose )
	fprintf(stderr, "%s: driver directory: %s\n", __progname, g.nv_drv_dir);
#if defined(__x86_64__) || defined(__aarch64__)
    strcpy(g.nv_drv32_dir, NV_DRV32_DIR_PREFIX);
    if( strlen(g.new_kmod_ver_str) > 0 ) {
	strcat(g.nv_drv32_dir, g.new_kmod_ver_str);
    } else {
	strcat(g.nv_drv32_dir, g.last_kmod_ver_str);
    }
    strcat(g.nv_drv32_dir, "/");
#else
    strcpy(g.nv_drv32_dir, g.nv_drv_dir);
#endif
    if( g.o_verbose )
	fprintf(stderr, "%s: driver32 directory: %s\n", __progname, g.nv_drv32_dir);
}

void restore_symlinks()
{
    restore_symlink(0, "nvidia_drv.so", XDRV_DIR, NV_ETCLIB_SYML_DIR);
    restore_symlink(-1, "libGLcore.so.1", XLIB_DIR, NV_ETCLIB_SYML_DIR);
    restore_symlink(-1, "libnvidia-tls.so.1", TLS_LIB_DIR, NV_ETCLIB_SYML_DIR);
    restore_symlink(-1, "libnvidia-cfg.so.1", XLIB_DIR, NV_ETCLIB_SYML_DIR);
    restore_symlink(-1, "libEGL_nvidia.so.0", XLIB_DIR, NV_ETCLIB_SYML_DIR);
    restore_symlink(-1, "libGLESv2_nvidia.so.2", XLIB_DIR, NV_ETCLIB_SYML_DIR);
    restore_symlink(-1, "libGLESv1_CM_nvidia.so.1", XLIB_DIR, NV_ETCLIB_SYML_DIR);
    restore_symlink(-1, "libGLX_nvidia.so.0", XLIB_DIR, NV_ETCLIB_SYML_DIR);
    /* restore symlinks from other packages */
    if( using_glvnd_scheme() == 0 ) {
	restore_symlink(1, "libglx.so", XEXT_DIR, X_ETCLIB_SYML_DIR);
	if( is_nvidia_plain_symlink(XLIB_DIR"libGL.so.1") )
    	    unlink_file(XLIB_DIR, "libGL.so.1");
	restore_symlink(1, "libGL.so.1", XLIB_DIR, X_ETCLIB_SYML_DIR);
	restore_symlink(0, "libGLESv2.so.2", XLIB_DIR, X_ETCLIB_SYML_DIR);
	restore_symlink(0, "libGLESv1_CM.so.1", XLIB_DIR, X_ETCLIB_SYML_DIR);
	restore_symlink(0, "libEGL.so.1", XLIB_DIR, X_ETCLIB_SYML_DIR);
	restore_symlink(0, "libGLdispatch.so.0", XLIB_DIR, X_ETCLIB_SYML_DIR);
	restore_symlink(0, "libGLX.so.0", XLIB_DIR, X_ETCLIB_SYML_DIR);
	/* restore_symlink(0, "libwfb.so", XMOD_DIR, X_ETCLIB_SYML_DIR); */
    }
#if defined(__x86_64__) || defined(__aarch64__)
	if( detected_libs_32() )
	{
	    int verbosity = 0;
	    if( is_nvidia_plain_symlink(XLIB32_DIR"libnvidia-tls.so.1") )
		unlink_file(XLIB32_DIR, "libnvidia-tls.so.1");
	    restore_symlink(-1, "libnvidia-tls.so.1", TLS_LIB32_DIR, NV_ETCLIB_SYML32_DIR);

	    if( is_nvidia_plain_symlink(XLIB32_DIR"libGLcore.so.1") )
		unlink_file(XLIB32_DIR, "libGLcore.so.1");
	    restore_symlink(-1, "libGLcore.so.1", XLIB32_DIR, NV_ETCLIB_SYML32_DIR);

	    if( is_nvidia_plain_symlink(XLIB32_DIR"libEGL_nvidia.so.0") )
		unlink_file(XLIB32_DIR, "libEGL_nvidia.so.0");
	    restore_symlink(-1, "libEGL_nvidia.so.0", XLIB32_DIR, NV_ETCLIB_SYML32_DIR);

	    if( is_nvidia_plain_symlink(XLIB32_DIR"libGLESv2_nvidia.so.2") )
		unlink_file(XLIB32_DIR, "libGLESv2_nvidia.so.2");
	    restore_symlink(-1, "libGLESv2_nvidia.so.2", XLIB32_DIR, NV_ETCLIB_SYML32_DIR);

	    if( is_nvidia_plain_symlink(XLIB32_DIR"libGLESv1_CM_nvidia.so.1") )
		unlink_file(XLIB32_DIR, "libGLESv1_CM_nvidia.so.1");
	    restore_symlink(-1, "libGLESv1_CM_nvidia.so.1", XLIB32_DIR, NV_ETCLIB_SYML32_DIR);

	    if( is_nvidia_plain_symlink(XLIB32_DIR"libGLX_nvidia.so.0") )
		unlink_file(XLIB32_DIR, "libGLX_nvidia.so.0");
	    restore_symlink(-1, "libGLX_nvidia.so.0", XLIB32_DIR, NV_ETCLIB_SYML32_DIR);

	    /* restore symlinks from other packages */
	    if( using_glvnd_scheme() == 0 ) {
		if( is_nvidia_plain_symlink(XLIB32_DIR"libGL.so.1") )
		    unlink_file(XLIB32_DIR, "libGL.so.1");
		else
		    verbosity = 1;
		restore_symlink(verbosity, "libGL.so.1", XLIB32_DIR, X_ETCLIB_SYML32_DIR);
		restore_symlink(0, "libGLESv2.so.2", XLIB32_DIR, X_ETCLIB_SYML32_DIR);
		restore_symlink(0, "libGLESv1_CM.so.1", XLIB32_DIR, X_ETCLIB_SYML32_DIR);
		restore_symlink(0, "libEGL.so.1", XLIB32_DIR, X_ETCLIB_SYML32_DIR);
		restore_symlink(0, "libGLdispatch.so.0", XLIB32_DIR, X_ETCLIB_SYML32_DIR);
		restore_symlink(0, "libGLX.so.0", XLIB32_DIR, X_ETCLIB_SYML32_DIR);
	    }
	}
#endif
}

int read_version_from_file(char *version, const char *path)
{
    int version_file;
    char newver[VER_LEN_MAX];
    char *newver_p = (char *)&newver;

    strcpy(version, "");
    version_file = open(path, O_RDONLY);
	if( version_file > 0 )
	{
	    int i = 0;
	    while( read(version_file, newver_p, 1) > 0 && i < VER_LEN_MAX )
	    {
		if( newver[i] == '\n' )
		{
		    newver[i] = '\0';
		    break;
		}
		newver_p++; i++;
	    }
	    newver[i] = '\0';
	    close( version_file );
	    strcpy(version, newver);
	    return 1;
	}
    return 0;
}

int read_version_from_proc(char *version, const char *path)
{
    /* string sample
	NVRM version: NVIDIA UNIX ARCH Kernel Module  340.24  Wed Jul  2 14:24:20 PDT 2014 */
    FILE *proc_file;
    if( !version )
	return 0;
    strcpy(version, "");
    if( !path )
	return 0;
    proc_file = fopen(path, "r");
    if( proc_file )
    {
	while( !feof(proc_file) )
	{
	    char buf[128];
	    char *line;
	    int len;
	    line = fgets(buf, 127, proc_file);
	    if( line && strlen(line) > 13 && strncmp("NVRM version:", line, 13) == 0 )
	    {
		int n = 0;
		char *start;
		len = strlen(line);
		start = strstr(line, "Module");
		if( start )
		{
		    int i;
		    n = len-strlen(start);
		    while( n < len )
		    {
			start++; n++;
			if( isdigit(buf[n]) != 0 ) break;
		    }
		    if( n >= len ) break;
		    for(i=0; i<VER_LEN_MAX; i++, n++)
		    {
			if( isspace(buf[n]) != 0 || n >= len ) break;
			version[i] = buf[n];
			if( version[i] == '-' ) version[i] = '.';
		    }
		    version[i] = '\0';
		    if( strlen(version) > 0 )
		    {
			if( g.o_verbose )
			    fprintf(stderr, "%s: loaded kernel module version - %s\n", __progname, version);
			return 1;
		    }
		}
		break;
	    }
	}
	fclose(proc_file);
    }
    return 0;
}

int write_version_to_file(const char *version, const char *path)
{
    int ret = 0;
    int version_file;
    int len;
    if( !version ) return ret;
    len = strlen(version);
    if( len <= 0 ) return ret;
    version_file = open(path, O_WRONLY);
    if( version_file > 0 )
    {
	if( write(version_file, version, len) > 0 )
	{
	    if( write(version_file, "\n", 1) > 0 )
		ret = 1;
	}
	close( version_file );
    }
    return ret;
}

int version_compare(const struct nvver ver1, const struct nvver ver2)
{
    int i;
    for(i = 0; i < VER_NUMBERS; i++)
    {
	if( ver1.num[i] > ver2.num[i] )
	    return 1;
	else if( ver1.num[i] < ver2.num[i] )
	    return -1;
    }
    return 0;
}

struct nvver string_to_version(const char *version)
{
    struct nvver res;
    int number = 0;
    char *strt;
    char *token;
    int i = 0;
    for(i=0; i < VER_NUMBERS; i++) res.num[i] = 0;
    strt = strdup(version);
    token = strtok(strt, ".");
    while( token )
    {
	res.num[number] = strtol(token, NULL, 10); break;
	if( res.num[number] < 0 ) res.num[number] = 0;
	number++;
	token = strtok(NULL, ".");
    }
    free(strt);
    return res;
}

int detected_libs_32()
{
    if( detected_libs_32_ >= 0 )
	return detected_libs_32_;

    if( access("/lib/libc.so.6", F_OK) == 0 )
	detected_libs_32_ = 1;
    else
	detected_libs_32_ = 0;

    return detected_libs_32_;
}

int is_nvidia_plain_symlink(const char *path)
{
    int ret = 0;
    char buf[128];
    int readlink_len = 0;

	memset(buf, 0, 128);
	readlink_len = readlink(path, buf, 128);
	if( readlink_len > 0 )
	{
	    if( readlink_len >= 128 ) readlink_len = 127;
	    buf[readlink_len] = '\0';
	    if( strstr(buf, "/nvidia_") )
		ret = 1;
	}
    return ret;
}

int is_exists_dir_file(const char *dir,const char *file)
{
    char dst[PATH_MAX];
    strcpy(dst, dir);
    strcat(dst, file);
    return access(dst, F_OK) == 0;
}

struct nvdevlist* new_nvidia_dev(struct nvdevlist **cards, unsigned int vnd, unsigned int dev, struct xinflist *xnf)
{
    struct nvdevlist *nvdev = malloc(sizeof(struct nvdevlist));
    nvdev->vendor = vnd;
    nvdev->device = dev;
    nvdev->xinf = xnf;
    if( cards ) {
	nvdev->next = *cards;
	*cards = nvdev;
    }
    return nvdev;
}

int using_glvnd_scheme()
{
#if GLVND_SCHEME > 0
    return 1;
#elif GLVND_SCHEME == 0
    if( __new_glvnd_scheme_detect_result < 0 ) {
	__new_glvnd_scheme_detect_result = access(XLIB_DIR"X11/libGLX.so.0.0.0", F_OK) != 0;
    }
    return __new_glvnd_scheme_detect_result;
#else
    return 0;
#endif
}
