#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <limits.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <glob.h>
#include <unistd.h>
#include <errno.h>
#include <sysfs/libsysfs.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <ctype.h>

#include "settings.h"
#include "lib.h"

#define XORG_NVIDIA_GLXMOD_CONF "/etc/X11/xorg.conf.d/09-nvidia.conf"
#define LD_SO_CONF_NVIDIA "/etc/ld.so.conf.d/nvidia.conf"
#define LD_SO_CONF_NVIDIA32 "/etc/ld.so.conf.d/nvidia32.conf"
#define NVIDIA_CURRENT_LIB_DIR "/etc/libnvidiacurrent"
#define NVIDIA_CURRENT_LIB32_DIR "/etc/libnvidia32current"
#define XORG_MATCHDRIVER_NVIDIA_VER "360.0"
#define XORG_GLVND_NVIDIA_VER "380.0"
#define XORG_PRIME_NVIDIA_VER "435.0"

extern const char *__progname;
struct envglobal g;
struct nvdevlist *g_nvidia_cards_database = 0;

void program_version( FILE *output )
{
    fprintf(output, PACKAGE" kernel module switcher v%s by Sergey Turchin <zerg@altlinux.org>\n", PACKAGE_VERSION);
}

void program_usage( FILE *output )
{
    program_version( output );
    fprintf(output,
"Usage:\n"
"  nvidia -h				-- display command usage\n"
"  nvidia -V				-- display program version\n"
"  nvidia -v, --verbose			-- enable verbose output\n"
);
}

int match_nvidia(unsigned int vendor_id, unsigned int device_id)
{
    struct nvdevlist *nvcard = 0;

    for(nvcard = g_nvidia_cards_database; nvcard; nvcard = nvcard->next)
    {
	if( nvcard->vendor == vendor_id && nvcard->device == device_id )
	{
	    return 1;
	}
    }
    return 0;
}

int main(int argc, char **argv)
{
    int is_kmod_change = 0;
    struct xinflist *paired_xinfs = 0;
    struct nvdevlist *nvidia_cards = 0;
    static struct option long_options[] = { 
	{ "verbose", 0, NULL, 'v' }, 
	{ "help", 0, NULL, 'h' }, 
	{ "version", 0, NULL, 'V'},
	{ NULL, 0, NULL, 0 } };

    init_global();

    while(1)
    {
	int c = getopt_long( argc, argv, "hvV", long_options, NULL );
	if( c == -1 ) 
    	    break;
	switch( c )
	{
	    case 'v':
		g.o_verbose = 1;
		break;
	    case 'V':
		program_version( stdout );
		exit(0);
	    case 'h':
		program_usage( stdout );
		exit(0);
	    default: 
		program_usage( stderr );
		exit(1);
	}
    }

    get_kernel_ver();
    get_last_kmod_ver();

    { /* find paired xinfs */
	glob_t globbuf;
	size_t i;

	if( glob(XINF_DIR"nvidia-*.xinf", GLOB_NOESCAPE, NULL, &globbuf) == 0 )
	{	
	    for(i=0; i < globbuf.gl_pathc; i++)
	    {
		const char *path = *(globbuf.gl_pathv);
		char *start;
		start = strstr(path, "-");
		if( start )
		{
			char *end;
			start++;
			end = strstr(start, ".xinf");
			if( end )
			{
			    int len = strlen(start)-strlen(end);
			    if( len < VER_LEN_MAX && len > 0 )
			    {
				char nv_kmod_file[PATH_MAX];
				char nv_ver_str[VER_LEN_MAX];
				strncpy(nv_ver_str, start, len);
				nv_ver_str[len] = '\0';
				strcpy(nv_kmod_file, NV_KMOD_DIR);
				strcat(nv_kmod_file, g.kernel_ver_str);
				strcat(nv_kmod_file, "-");
				strcat(nv_kmod_file, nv_ver_str);
				if( access(nv_kmod_file, F_OK) == 0 )
				{
				    struct xinflist *xinf = malloc(sizeof(struct xinflist));
				    strcpy(xinf->xinf, path);
				    strcpy(xinf->module, nv_kmod_file);
				    strcpy(xinf->version, nv_ver_str);
				    xinf->supported = 0;
				    xinf->next = paired_xinfs;
				    paired_xinfs = xinf;
				    if( g.o_verbose )
					fprintf(stderr, "%s: found %s\n", __progname, path);
				}
			    }
			}
		}
		globbuf.gl_pathv++;
	    }
	}
    }

    /* fill nvidia cards database */
    if( paired_xinfs )
    {
	struct xinflist *xinf;
	for(xinf = paired_xinfs; xinf; xinf = xinf->next)
	{
		unsigned int vendor = 0;
		unsigned int device = 0;
		FILE *xinffile;
		xinffile = fopen(xinf->xinf, "r");
		if( xinffile )
		{
		    while( !feof(xinffile) )
		    {
			char buf[LINE_MAX];
			char *line;
			int len;
			line = fgets(buf, LINE_MAX-1, xinffile);
			if( line )
			{
			    len = strlen(line);
			    if( buf[len-1] == '\n' )
			    {
				len--;
				buf[len] = '\0';
			    }
			    vendor = 0;
			    device = 0;
			    sscanf(line, "alias pcivideo:v0000%04xd0000%04xsv", &vendor, &device);
			    if( vendor > 0 && device > 0 )
			    {
				new_nvidia_dev(&g_nvidia_cards_database, vendor, device, xinf);
			    }
			}
		    }
		    fclose(xinffile);
		}
	}
    }

    { /* scan pci */
	g.other_videocards = 0;
	g.intel_videocards = 0;
	g.nvidia_videocards = 0;
	g.nvidia_3dcontrollers = 0;
	struct sysfs_bus *bus;
	bus = sysfs_open_bus("pci");
	if( bus != NULL )
	{
	    struct dlist *devlist;
	    struct sysfs_device *curdev;
	    devlist = sysfs_get_bus_devices(bus);
	    if( devlist )
	    {
		dlist_for_each_data(devlist, curdev, struct sysfs_device)
		{
		    struct sysfs_attribute *class_attr;
		    class_attr = sysfs_get_device_attr(curdev, "class");
		    if( class_attr != NULL )
		    {
			long int class_id;
			class_id = strtol(class_attr->value,NULL,16);
			if( class_id == 0x030000 || class_id == 0x030200 )
			{
			    struct sysfs_attribute *vendor_attr;
			    struct sysfs_attribute *device_attr;
			    vendor_attr = sysfs_get_device_attr(curdev, "vendor");
			    device_attr = sysfs_get_device_attr(curdev, "device");
			    if( vendor_attr != NULL && device_attr != NULL )
			    {
				unsigned int vendor_id, device_id;
				vendor_id = strtol(vendor_attr->value,NULL,16);
				device_id = strtol(device_attr->value,NULL,16);
				if( vendor_id == 0x10de
				    || vendor_id == 0x12d2
				    || match_nvidia(vendor_id, device_id) )
				{
				    new_nvidia_dev(&nvidia_cards, vendor_id, device_id, 0);
				    if( class_id == 0x030000 ) {
					g.nvidia_videocards++;
					if( g.o_verbose )
					    fprintf(stderr, "%s: found NVIDIA VGA controller %08x:%08x\n", __progname, vendor_id, device_id);
				    } else if( class_id == 0x030200) {
					g.nvidia_3dcontrollers++;
					if( g.o_verbose )
					    fprintf(stderr, "%s: found NVIDIA 3D controller %08x:%08x\n", __progname, vendor_id, device_id);
				    }
				}
				else
				{
				    if( class_id == 0x030000 ) {
					if( vendor_id == 0x8086 ) {
					    g.intel_videocards++;
					} else {
					    g.other_videocards++;
					}
					if( g.o_verbose )
					    fprintf(stderr, "%s: found VGA controller %08x:%08x\n", __progname, vendor_id, device_id);
				    }
				}
			    }
			    else
			    {
				if( g.o_verbose )
				    fprintf(stderr, "%s: found unrecognized 0x03 class device\n", __progname);
			    }
			}
		    }
		}
	    }
	}
	else
	    fprintf(stderr,"%s: failed to scan pci via sysfs\n", __progname);
    }

    /* match xinfs */
    if( paired_xinfs && nvidia_cards )
    {
	struct xinflist *xinf;
	for(xinf = paired_xinfs; xinf; xinf = xinf->next)
	{
	    struct nvdevlist *nvcard = 0;
	    for(nvcard = g_nvidia_cards_database; nvcard; nvcard = nvcard->next)
	    {
		struct nvdevlist *nvdev;
		for( nvdev = nvidia_cards; nvdev; nvdev=nvdev->next )
		{
		    if( nvdev->vendor == nvcard->vendor
			&& nvdev->device == nvcard->device
			&& xinf == nvcard->xinf )
		    {
			xinf->supported++;
		    }
		}
	    }
	}
    }

    /* select best version */    
    if( paired_xinfs && nvidia_cards )
    {
	int max_support = 0;
	struct xinflist *xinf;
	for(xinf = paired_xinfs; xinf; xinf = xinf->next)
	{
	    if( xinf->supported > 0
		&& xinf->supported >= max_support
		&& version_compare(string_to_version(xinf->version), string_to_version(g.new_kmod_ver_str)) > 0 )
	    {
		max_support = xinf->supported;
		strcpy(g.new_kmod_ver_str, xinf->version);
	    }
	}
	if( strlen(g.new_kmod_ver_str) > 0 )
	{
	    if( g.o_verbose )
		printf("%s: best kernel module version - %s\n", __progname, g.new_kmod_ver_str);
	}
    }

    /* unload kernel module */
    if( strlen(g.new_kmod_ver_str) > 0 )
    {
	int sysret = 0;
	{ /* read loaded module version */
	    FILE *proc_modules;
	    proc_modules = fopen("/proc/modules", "r");
	    if( proc_modules )
	    {
		while( !feof(proc_modules) )
		{
		    char buf[256];
		    char *line;
		    line = fgets(buf, 255, proc_modules);
		    if( line && strlen(line) > 6 && strncmp("nvidia", line, 6) == 0 && isspace(buf[6]) )
		    {
			read_version_from_proc(g.proc_kmod_ver_str, "/proc/driver/nvidia/version");
		    }
		}
		fclose(proc_modules);
	    }
	}
	if( strlen(g.new_kmod_ver_str) > 0
	    && strlen(g.proc_kmod_ver_str) > 0
	    && strcmp(g.new_kmod_ver_str, g.proc_kmod_ver_str) != 0 )
	{
	    sysret = system("/sbin/modprobe -r nvidia");
	    if( sysret != 0 )
	    {
	        fprintf(stderr, "%s: Warning! You must reboot computer or shutdown X-server and unload kernel module by command `/sbin/modprobe -r nvidia`\n", __progname);
	    }
	    else
	    {
	        if( g.o_verbose )
		    fprintf(stderr, "%s: unload kernel module\n", __progname);
	    }
	}
	else
	{
	    if( g.o_verbose )
	        fprintf(stderr, "%s: don't unload kernel module\n", __progname);
	}
    }

    /* switch kernel module */
    if( strlen(g.new_kmod_ver_str) > 0 )
    {
	char kmod_dir[PATH_MAX];
	char nv_kmod_file[PATH_MAX];
	char nv_modesetmod_file[PATH_MAX];
	char nv_modesetmod_path[PATH_MAX];
	char nv_uvmmod_file[PATH_MAX];
	char nv_uvmmod_path[PATH_MAX];
	char nv_drmmod_file[PATH_MAX];
	char nv_drmmod_path[PATH_MAX];
	char version_file_path[PATH_MAX];
	int depmod_sysret = 0;

	is_kmod_change = strcmp(g.last_kmod_ver_str, g.new_kmod_ver_str) != 0;

	strcpy(kmod_dir, K_MOD_PREFIX); strcat(kmod_dir, g.kernel_ver_str); strcat(kmod_dir, "/nVidia/");

	strcpy(nv_kmod_file, g.kernel_ver_str); strcat(nv_kmod_file, "-"); strcat(nv_kmod_file, g.new_kmod_ver_str);

	strcpy(nv_modesetmod_file, "modeset-"); strcat(nv_modesetmod_file, g.kernel_ver_str); strcat(nv_modesetmod_file, "-"); strcat(nv_modesetmod_file, g.new_kmod_ver_str);
	strcpy(nv_modesetmod_path, NV_KMOD_DIR); strcat(nv_modesetmod_path, nv_modesetmod_file);

	strcpy(nv_uvmmod_file, "uvm-"); strcat(nv_uvmmod_file, g.kernel_ver_str); strcat(nv_uvmmod_file, "-"); strcat(nv_uvmmod_file, g.new_kmod_ver_str);
	strcpy(nv_uvmmod_path, NV_KMOD_DIR); strcat(nv_uvmmod_path, nv_uvmmod_file);

	strcpy(nv_drmmod_file, "drm-"); strcat(nv_drmmod_file, g.kernel_ver_str); strcat(nv_drmmod_file, "-"); strcat(nv_drmmod_file, g.new_kmod_ver_str);
	strcpy(nv_drmmod_path, NV_KMOD_DIR); strcat(nv_drmmod_path, nv_drmmod_file);

	if( is_kmod_change )
	    printf("%s: switching kernel module from %s to %s\n", __progname, g.last_kmod_ver_str, g.new_kmod_ver_str);
	mk_symlink_from_to(0, nv_kmod_file, NV_KMOD_DIR, "nvidia.ko", kmod_dir);
	if( access(nv_modesetmod_path, F_OK) == 0 )
	    mk_symlink_from_to(0, nv_modesetmod_file, NV_KMOD_DIR, "nvidia-modeset.ko", kmod_dir);
	else
	    unlink_file(kmod_dir, "nvidia-modeset.ko");
	if( access(nv_uvmmod_path, F_OK) == 0 )
	    mk_symlink_from_to(0, nv_uvmmod_file, NV_KMOD_DIR, "nvidia-uvm.ko", kmod_dir);
	else
	    unlink_file(kmod_dir, "nvidia-uvm.ko");
	if( access(nv_drmmod_path, F_OK) == 0 )
	    mk_symlink_from_to(0, nv_drmmod_file, NV_KMOD_DIR, "nvidia-drm.ko", kmod_dir);
	else
	    unlink_file(kmod_dir, "nvidia-drm.ko");

	/* fill version text file */
	strcpy(version_file_path, K_MOD_PREFIX);
	strcat(version_file_path, g.kernel_ver_str);
	strcat(version_file_path, K_MOD_VER_SUFFIX);
	write_version_to_file(g.new_kmod_ver_str, version_file_path);

	if( is_kmod_change )
	    depmod_sysret = system("/sbin/depmod -a");
	if( depmod_sysret != 0 )
	{
	    /* something */
	}
    }

    get_nv_drv_dir();

    /* switch current nvidia libs directory */
    if( strlen(g.nv_drv_dir) > 0 )
    {
	char src[PATH_MAX];
	char dst[PATH_MAX];
	strcpy(src, NVIDIA_CURRENT_LIB_DIR);
	strcpy(dst, g.nv_drv_dir);
	unlink(src);
	if( symlink(dst, src) != 0 )
	{
	    if( g.o_verbose )
		fprintf(stderr, "%s: failed make symlink %s (to %s): %s\n", __progname, src, dst, strerror(errno));
	}
#if defined(__x86_64__) || defined(__aarch64__)
	if( detected_libs_32() && strlen(g.nv_drv32_dir) > 0 )
	{
	    strcpy(src, NVIDIA_CURRENT_LIB32_DIR);
	    strcpy(dst, g.nv_drv32_dir);
	    unlink(src);
	    if( symlink(dst, src) != 0 )
	    {
		if( g.o_verbose )
		    fprintf(stderr, "%s: failed make symlink %s (to %s): %s\n", __progname, src, dst, strerror(errno));
	    }
	}
#endif
    }

    if( (g.nvidia_videocards > 0 || g.nvidia_3dcontrollers > 0) && strlen(g.new_kmod_ver_str) > 0 ) {
	int nouveau_loaded = 0;
	{ /* check for nouveau kernel module */
	    FILE *proc_modules;
	    proc_modules = fopen("/proc/modules", "r");
	    if( proc_modules )
	    {
		while( !feof(proc_modules) )
		{
		    char buf[256];
		    char *line;
		    line = fgets(buf, 255, proc_modules);
		    if( line && strlen(line) > 7 && strncmp("nouveau", line, 7) == 0 && isspace(buf[7]) )
		    {
			nouveau_loaded = 1;
		    }
		}
		fclose(proc_modules);
	    }
	}
	if( !nouveau_loaded ) {
	    int sysret = 0;
	    restore_symlinks();
	    if( g.nvidia_videocards > 0
		&& version_compare( string_to_version(g.new_kmod_ver_str), string_to_version(XORG_MATCHDRIVER_NVIDIA_VER)) < 0 ) {
		sysret = system("/sbin/modprobe nvidia");
	    }
	    /* add only old non-glvnd nvidia libraries to ld.so.conf */
	    if( using_glvnd_scheme() == 1 ) {
		if( version_compare( string_to_version(g.new_kmod_ver_str), string_to_version(XORG_GLVND_NVIDIA_VER)) < 0 ) {
		    {
			FILE *fl = fopen(LD_SO_CONF_NVIDIA, "w");
			if( fl ) {
			    fprintf(fl, NVIDIA_CURRENT_LIB_DIR"\n");
			    fclose(fl);
			}
		    }
#if defined(__x86_64__) || defined(__aarch64__)
		    {
			FILE *fl = fopen(LD_SO_CONF_NVIDIA32, "w");
			if( fl ) {
			    fprintf(fl, NVIDIA_CURRENT_LIB32_DIR"\n");
			    fclose(fl);
			}
		    }
#endif
		} else {
		    if( access(LD_SO_CONF_NVIDIA, F_OK) == 0 ) {
			FILE *fl = fopen(LD_SO_CONF_NVIDIA, "w");
			if( fl ) {
			    fclose(fl);
			}
		    }
#if defined(__x86_64__) || defined(__aarch64__)
		    if( access(LD_SO_CONF_NVIDIA32, F_OK) == 0 ) {
			FILE *fl = fopen(LD_SO_CONF_NVIDIA32, "w");
			if( fl ) {
			    fclose(fl);
			}
		    }
#endif
		}
	    }

	    /* allow load nvidia modules by xorg */
	    FILE *fl = fopen(XORG_NVIDIA_GLXMOD_CONF, "w");
	    if( fl != NULL ) {
		fprintf(fl, "# Autogenerated by drvpre.d/nvidia\n");
		fprintf(fl, "Section \"Files\"\n");
		fprintf(fl, "    ModulePath \"/etc/libnvidiacurrent\"\n");
		fprintf(fl, "    ModulePath \""XLIB_DIR"X11/modules\"\n");
		fprintf(fl, "EndSection\n");
		/* allow to load nvidia xorg driver */
		if( version_compare( string_to_version(g.new_kmod_ver_str), string_to_version(XORG_MATCHDRIVER_NVIDIA_VER)) >= 0 ) {
		    fprintf(fl, "\n");
		    fprintf(fl, "Section \"OutputClass\"\n");
		    fprintf(fl, "    Identifier  \"nvidia\"\n");
		    fprintf(fl, "    MatchDriver \"nvidia-drm\"\n");
		    fprintf(fl, "    Driver      \"nvidia\"\n");
		    if( using_glvnd_scheme() == 1 ) {
		        fprintf(fl, "    Option      \"AllowEmptyInitialConfiguration\"\n");
		    }
		    fprintf(fl, "EndSection\n");
		}
		if( version_compare( string_to_version(g.new_kmod_ver_str), string_to_version(XORG_PRIME_NVIDIA_VER)) >= 0 ) {
		    fprintf(fl, "\n");
		    fprintf(fl, "Section \"ServerLayout\"\n");
		    fprintf(fl, "    Identifier  \"nvidia\"\n");
		    fprintf(fl, "    Option      \"AllowNVIDIAGPUScreens\"\n");
		    fprintf(fl, "EndSection\n");
		}
		fclose(fl);
	    }
	    if( using_glvnd_scheme() == 0
		|| version_compare( string_to_version(g.new_kmod_ver_str), string_to_version(XORG_MATCHDRIVER_NVIDIA_VER)) < 0 ) {
		if( access("/etc/X11/xorg.conf", F_OK) != 0 && access("/etc/X11/xorg.conf.d/10-monitor.conf", F_OK) != 0 ) {
		    sysret = system("xsetup-monitor --create-only --driver nvidia");
		    if( sysret != 0 ) {
			fprintf(stderr, "%s: warning! xsetup-monitor failed with code %d.\n", __progname, sysret);
		    }
		}
	    }
	    /* switch nvidia libraries */
	    if( using_glvnd_scheme() == 1 || (g.intel_videocards <= 0 && g.other_videocards <= 0) ) {
		sysret = system(g.o_verbose? "/usr/libexec/X11/drv.d/nvidia -v": "/usr/libexec/X11/drv.d/nvidia");
		if( is_kmod_change ) {
		    sysret = system("/sbin/ldconfig");
		}
	    }
	}
    } else {
	/* clean ld.so.conf.d/nvidia.conf */
	if( access(LD_SO_CONF_NVIDIA, F_OK) == 0 ) {
	    FILE *fl = fopen(LD_SO_CONF_NVIDIA, "w");
	    if( fl ) {
	        fclose(fl);
	    }
	}
#if defined(__x86_64__) || defined(__aarch64__)
	if( access(LD_SO_CONF_NVIDIA32, F_OK) == 0 ) {
	    FILE *fl = fopen(LD_SO_CONF_NVIDIA32, "w");
	    if( fl ) {
	        fclose(fl);
	    }
	}
#endif
	/* clean xorg.conf.d/09-nvidia.conf */
	if( access(XORG_NVIDIA_GLXMOD_CONF, F_OK) == 0 ) {
	    FILE *fl = fopen(XORG_NVIDIA_GLXMOD_CONF, "w");
	    if( fl ) {
	        fclose(fl);
	    }
	}
    }

    /* restore_symlinks(); */

/* used with old scheme
    if( mk_symlink(0, "nvidia.xinf", NV_ETCLIB_SYML_DIR) == 0 )
	mk_symlink_from_to(0, "null", "/dev/", "nvidia.xinf", NV_ETCLIB_SYML_DIR);
*/
    return(0);
}
